!*******************************************************
! fftw3ds.f90 - cubic decomposition subroutines for fftw

  subroutine mpi_initialize
!! initialize mpi parameters
    implicit none
    include 'mpif.h'
    include 'fftw3d.fh'

    integer(4) :: i, j, key, subset, nodes_returned
    integer(4) :: dims(3), ndim
    logical :: periodic(3), reorder

!! set up global mpi communicator

    call mpi_init(ierr)
    if (ierr /= mpi_success) call mpi_abort(mpi_comm_world,ierr,ierr)

    call mpi_comm_size(mpi_comm_world,nodes_returned,ierr)
    if (ierr /= mpi_success) call mpi_abort(mpi_comm_world,ierr,ierr)
    if (nodes_returned /= nodes + nodes * nt) then
      write(*,*) 'fftw3d compiled for a different number of nodes'
      write(*,*) 'mpirun nodes=',nodes_returned,'fftw3d nodes=', &
                  nodes + nodes * nt
      call mpi_abort(mpi_comm_world,ierr,ierr)
    endif
    call mpi_comm_rank(mpi_comm_world,rank,ierr)
    if (ierr /= mpi_success) call mpi_abort(mpi_comm_world,ierr,ierr)

!! split global communicator into fftw + cube subsets

    subset = 0
    key = subset
    call mpi_comm_split(mpi_comm_world,subset,key,mpi_comm_cube,ierr)
    if (ierr /= mpi_success) call mpi_abort(mpi_comm_world,ierr,ierr)
    call mpi_comm_size(mpi_comm_cube, cube_procs, ierr)
    if (ierr /= mpi_success) call mpi_abort(mpi_comm_world,ierr,ierr)
    if (cube_procs /= nodes) then
      write(*,*) 'mpirun program nodes=',cube_procs,'fftw3d slab nodes=',nodes
      call mpi_abort(mpi_comm_world,ierr,ierr)
    endif
    call mpi_comm_rank(mpi_comm_cube, cube_rank, ierr)
    if (ierr /= mpi_success) call mpi_abort(mpi_comm_world,ierr,ierr)

!! calculate coordinates within slab for cube processes

    slab_coord(3) = cube_rank / nodes_slab
    slab_rank = cube_rank - slab_coord(3) * nodes_slab
    slab_coord(2) = slab_rank / nodes_dim
    slab_coord(1) = slab_rank - slab_coord(2) * nodes_dim
    do j = 0, nodes_dim - 1
      do i = 0, nodes_dim - 1
        slab_neighbor(i,j) = i + j * nodes_dim + slab_coord(3) &
                           * nodes_slab
      enddo
    enddo

!! create cartesian communicator based on cubic decomposition

    dims(:) = nodes_dim
    periodic(:) = .true.
    reorder = .false.
    ndim = 3

    call mpi_cart_create(mpi_comm_cube, ndim,dims, periodic, &
                       reorder, mpi_comm_cart, ierr)
    call mpi_comm_rank(mpi_comm_cart, cart_rank, ierr)
    call mpi_cart_coords(mpi_comm_cart, cart_rank, ndim,  &
                         cart_coords, ierr)
    do i = 0, ndim-1
      call mpi_cart_shift(mpi_comm_cart, i, 1, cart_neighbor(2*(i+1)-1), &
                          cart_neighbor(2*(i+1)), ierr)
    enddo

  end subroutine mpi_initialize

  subroutine pack_slab
!! pack cubic data into slab decomposition for fftw transform
    implicit none
    include 'mpif.h'
    include 'fftw3d.fh'

    integer(4) :: i,j,k,i0,j0,i1,j1,k1
    integer(4) :: slab_slice,num_elements,tag,rtag

    integer(4), dimension(2*nodes_dim**2) :: requests
    integer(4), dimension(MPI_STATUS_SIZE,2*nodes_dim**2) :: wait_status

    num_elements = nc_node_dim * nc_node_dim * nc_slab

!! swap data

    do j = 0, nodes_dim - 1
      do i = 0, nodes_dim - 1
        slab_slice = i + j * nodes_dim
        tag = rank**2
        rtag= slab_neighbor(i,j)**2
        call mpi_isend(cube_data(1,1,slab_slice*nc_slab + 1), num_elements, &
                       mpi_real, slab_neighbor(i,j), tag, mpi_comm_cube, &
                       requests(slab_slice+1),ierr)
        call mpi_irecv(recv_cube(1,1,1,slab_slice), &
                       num_elements, mpi_real, slab_neighbor(i,j),rtag, &
                       mpi_comm_cube, requests(slab_slice+1+nodes_dim**2), &
                       ierr)
      enddo
    enddo

    call mpi_waitall(2*nodes_dim**2, requests, wait_status, ierr)

!! place data in the slab

    do j = 0, nodes_dim - 1
      j0 = j * nc_node_dim + 1
      j1 = (j + 1) * nc_node_dim
      do i = 0, nodes_dim - 1
        i0 = i * nc_node_dim + 1
        i1 = (i + 1) * nc_node_dim
        slab_slice = i + j * nodes_dim
        slab(i0:i1,j0:j1,:) = recv_cube(:,:,:,slab_slice)
      enddo
    enddo

  end subroutine pack_slab

  subroutine unpack_slab 
!! unpack slab data into cubic decomposition following fftw transform
    implicit none
    include 'mpif.h'
    include 'fftw3d.fh'
    
    integer(4) :: i,j,k,i0,j0,i1,j1,k1
    integer(4) :: slab_slice,num_elements,tag,rtag
    integer(4), dimension(2*nodes_dim**2) :: requests
    integer(4), dimension(MPI_STATUS_SIZE,2*nodes_dim**2) :: wait_status

!! place data in the recv_cube buffer

    do j = 0, nodes_dim - 1
      j0 = j * nc_node_dim + 1
      j1 = (j + 1) * nc_node_dim
      do i = 0, nodes_dim - 1
        i0 = i * nc_node_dim + 1
        i1 = (i + 1) * nc_node_dim
        slab_slice = i + j * nodes_dim
        recv_cube(:,:,:,slab_slice) = slab(i0:i1,j0:j1,:)
      enddo
    enddo

    num_elements = nc_node_dim * nc_node_dim * nc_slab
 
!! swap data

   do j = 0, nodes_dim - 1
      do i = 0, nodes_dim - 1
        slab_slice = i + j * nodes_dim
        tag  = rank**2
        rtag = slab_neighbor(i,j)**2
        call mpi_isend(recv_cube(1,1,1,slab_slice), num_elements, &
                       mpi_real, slab_neighbor(i,j), tag, mpi_comm_cube, &
                       requests(slab_slice+1),ierr)
        call mpi_irecv(cube_data(1,1,slab_slice * nc_slab +1), &
                       num_elements, mpi_real, slab_neighbor(i,j), rtag, &
                       mpi_comm_cube, requests(slab_slice+1+nodes_dim**2), &
                       ierr)
      enddo
    enddo

    call mpi_waitall(2*nodes_dim**2,requests, wait_status, ierr)

  end subroutine unpack_slab

  subroutine fftw(command)
!! calculate fftw transform
    implicit none
    include 'mpif.h'
    include 'fftw3d.fh'

    integer(4) :: i
    integer(4) :: command, num_elements, fftw_proc, slab_low, slab_high,tag
    integer(4), dimension(nt) :: request
#ifdef STEST
    integer(4), dimension(mpi_status_size) :: status
#else
    integer(4), dimension(mpi_status_size,nt) :: status
#endif
    logical :: comp_flag

!! start fftw by sending command to appropriate processors
!! 0 ends fftw subprogram, 1 starts forward fft, -1 starts
!! backwards

    do i = 0, nt - 1
      fftw_proc = cube_procs -1 + (rank + 1) * nt - 1 + i
      tag = fftw_proc
      call mpi_isend(command,1,mpi_integer,fftw_proc,tag, &
                     mpi_comm_world, request(i+1), ierr)
    enddo

#ifdef STEST 
399 call system('sleep 1')
    do i=1,nt
      call mpi_test( request(i), comp_flag,status,ierr)
      if (.not. comp_flag) goto 399
    enddo
#else
    call mpi_waitall(nt,request,status,ierr)
#endif

!! if we are doing fft, send slab data

    if (command /= 0) then

!! call pack routine if we are going forward
  
      if (command > 0) call pack_slab

      num_elements = (nc_dim + 2) * nc_dim * nc_slab / nt

      do i = 0, nt - 1
        fftw_proc = cube_procs - 1 + (rank + 1) * nt - 1 + i
        slab_low = 1 + i * nc_slab / nt
        slab_high = (i + 1) * nc_slab / nt
        tag = fftw_proc
        call mpi_isend(slab(:,:,slab_low:slab_high), num_elements, mpi_real, &
                     fftw_proc, tag, mpi_comm_world, request(i+1), ierr)
      enddo

#ifdef STEST 
 400  call system('sleep 1')
      do i=1,nt
        call mpi_test( request(i), comp_flag, status, ierr)
        if (.not. comp_flag) goto 400
      enddo
#else
      call mpi_waitall(nt,request,status,ierr)
#endif

!! now receive the transformed answer

      do i = 0, nt - 1
        fftw_proc = cube_procs - 1 + (rank + 1) * nt - 1 + i
        slab_low = 1 + i * nc_slab / nt
        slab_high = (i + 1) * nc_slab / nt
        tag = fftw_proc
        call mpi_irecv(slab(:,:,slab_low:slab_high), num_elements, mpi_real, &
                       fftw_proc, tag, mpi_comm_world, request(i+1), ierr)
      enddo

#ifdef STEST
 401  call system('sleep 1')
      do i=1,nt
        call mpi_test( request(i), comp_flag, status, ierr)
        if (.not. comp_flag) goto 401
      enddo
#else 
      call mpi_waitall(nt,request,status,ierr)
#endif
  
!! unpack the slab data
 
      if (command < 0) call unpack_slab
 
    endif
  
  end subroutine fftw


